package ctbrec.recorder.download.hls;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.Encoding;
import com.iheartradio.m3u8.Format;
import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.PlaylistException;
import com.iheartradio.m3u8.PlaylistWriter;
import com.iheartradio.m3u8.data.MediaPlaylist;
import com.iheartradio.m3u8.data.Playlist;
import com.iheartradio.m3u8.data.PlaylistType;
import com.iheartradio.m3u8.data.TrackData;
import com.iheartradio.m3u8.data.TrackInfo;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.Recording;
import ctbrec.io.HttpClient;
import ctbrec.io.IoUtils;
import ctbrec.recorder.download.hls.SegmentPlaylist.Segment;

public class HlsDownload extends AbstractHlsDownload {

    private static final Logger LOG = LoggerFactory.getLogger(HlsDownload.class);

    protected Path downloadDir;

    private Queue<Future<SegmentDownload>> segmentDownloads = new LinkedList<>();

    private List<TrackData> segments = new LinkedList<>();

    private float targetDuration;

    public HlsDownload(HttpClient client) {
        super(client);
    }

    @Override
    public void init(Config config, Model model, Instant startTime, ExecutorService executorService) throws IOException {
        super.init(config, model, startTime, executorService);
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(Config.RECORDING_DATE_FORMAT);
        String formattedStartTime = formatter.format(ZonedDateTime.ofInstant(this.startTime, ZoneId.systemDefault()));
        Path modelDir = FileSystems.getDefault().getPath(config.getSettings().recordingsDir, model.getSanitizedNamed());
        downloadDir = FileSystems.getDefault().getPath(modelDir.toString(), formattedStartTime);
        createTargetDirectory();
    }

    @Override
    public AbstractHlsDownload call() throws Exception {
        super.call();
        updatePlaylist();
        return this;
    }

    @Override
    protected SegmentPlaylist getNextSegments(String segmentPlaylistUrl) throws IOException, ParseException, PlaylistException {
        SegmentPlaylist segmentPlaylist = super.getNextSegments(segmentPlaylistUrl);
        targetDuration = segmentPlaylist.targetDuration;
        return segmentPlaylist;
    }

    private void updatePlaylist() {
        downloadExecutor.submit(() -> {
            addNewSegmentsToPlaylist();
            try {
                MediaPlaylist playlist = new MediaPlaylist.Builder()
                        .withPlaylistType(PlaylistType.VOD)
                        .withMediaSequenceNumber(0)
                        .withTargetDuration(Math.round(targetDuration))
                        .withTracks(segments)
                        .build();

                // create a master playlist containing the media playlist
                Playlist master = new Playlist.Builder()
                        .withCompatibilityVersion(4)
                        .withExtended(true)
                        .withMediaPlaylist(playlist)
                        .build();

                // write the playlist to a file
                File output = new File(getTarget(), "playlist.m3u8");
                try (FileOutputStream fos = new FileOutputStream(output)) {
                    PlaylistWriter writer = new PlaylistWriter.Builder()
                            .withFormat(Format.EXT_M3U)
                            .withEncoding(Encoding.UTF_8)
                            .withOutputStream(fos)
                            .build();
                    writer.write(master);
                }
            } catch (IOException | ParseException | PlaylistException e) {
                LOG.error("Updating segment playlist failed", e);
            }
            LOG.trace("Segment queue size for {}: {}", model, segmentDownloads.size());
        });
    }

    private void addNewSegmentsToPlaylist() {
        Future<SegmentDownload> future;
        while ((future = segmentDownloads.peek()) != null && !Thread.currentThread().isInterrupted()) {
            try {
                if (running && future.isDone()) {
                    segmentDownloads.poll(); // future is done remove from queue
                    SegmentDownload segmentDownload = future.get();
                    segments.add(toTrack(segmentDownload.getSegment()));
                } else {
                    // first download in queue not finished, let's continue with other stuff
                    break;
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } catch (Exception e) {
                LOG.error("Segment download failed for model {}", model, e);
            }
        }
    }

    private TrackData toTrack(Segment segment) {
        String filename = segment.targetFile.getName();
        return new TrackData.Builder()
                .withUri(filename)
                .withTrackInfo(new TrackInfo(segment.duration, filename))
                .build();
    }

    protected void createTargetDirectory() throws IOException {
        if (!downloadDir.toFile().exists()) {
            Files.createDirectories(downloadDir);
        }
    }

    @Override
    public void finalizeDownload() {
        LOG.debug("Download for {} terminated", model);
    }

    @Override
    public void postprocess(Recording recording) {
        // nothing to do
    }

    @Override
    public void stop() {
        LOG.debug("Recording stopped");
        internalStop();
    }

    @Override
    protected void internalStop() {
        running = false;
    }

    @Override
    public File getTarget() {
        return downloadDir.toFile();
    }

    @Override
    public String getPath(Model model) {
        String absolutePath = getTarget().getAbsolutePath();
        String recordingsDir = Config.getInstance().getSettings().recordingsDir;
        String relativePath = absolutePath.replaceFirst(Pattern.quote(recordingsDir), "");
        return relativePath;
    }

    @Override
    public boolean isSingleFile() {
        return false;
    }

    @Override
    public long getSizeInByte() {
        return IoUtils.getDirectorySize(getTarget());
    }

    @Override
    protected OutputStream getSegmentOutputStream(Segment segment) throws FileNotFoundException, MalformedURLException {
        URL segmentUrl = new URL(segment.url);
        File tmp = new File(segmentUrl.getFile());
        String prefixedFileName = segment.prefix + '_' + tmp.getName();
        int questionMarkPosition = prefixedFileName.indexOf('?');
        if (questionMarkPosition > 0) {
            prefixedFileName = prefixedFileName.substring(0, questionMarkPosition);
        }
        if (!prefixedFileName.endsWith(".ts")) {
            prefixedFileName += ".ts";
        }
        segment.targetFile = FileSystems.getDefault().getPath(downloadDir.toAbsolutePath().toString(), prefixedFileName).toFile();
        return new FileOutputStream(segment.targetFile);
    }

    @Override
    protected void segmentDownloadFinished(SegmentDownload segmentDownload) {
        super.segmentDownloadFinished(segmentDownload);
        IoUtils.close(segmentDownload.getOutputStream(), "Couldn't close segment file");
    }

    @Override
    protected void execute(SegmentDownload segmentDownload) {
        segmentDownloads.add(segmentDownloadService.submit(segmentDownload));
    }
}
